"""Client for Andesite's HTTP routes.

There are multiple client classes in this module.
If you just want to use Andesite's HTTP endpoints,
use `HTTP`.

`HTTPInterface` contains the implementation
of the endpoint methods. It's an abstract base class,
if you want to inherit its methods you need to implement
`AbstractHTTP`.

Finally there is `HTTPBase` which is just the default
implementation of `AbstractHTTP`. `HTTP` is
just a combination of `HTTPBase` and `HTTPInterface`.


Attributes:
    USER_AGENT (str): User agent used by the `HTTP` client.
    SearcherType (Union[Searcher, str]): (Type alias) Types supported by `get_searcher`
"""

import abc
import logging
from enum import Enum
from typing import Any, Dict, Iterable, List, Optional, Union

import aiohttp
import yarl

import andesite
from .transform import build_from_raw, seq_build_all_items_from_raw

__all__ = ["USER_AGENT", "HTTPError",
           "Searcher", "SearcherType", "get_searcher",
           "AbstractHTTP", "HTTPInterface",
           "HTTPBase", "HTTP"]

log = logging.getLogger(__name__)

USER_AGENT = f"andesite.py/{andesite.__version__} (https://github.com/gieseladev/andesite.py)"


class HTTPError(Exception):
    """Andesite error.

    Attributes:
        code (int): HTTP error code
        message (str): Message sent by Andesite
    """
    __slots__ = ("code", "message")

    code: int
    message: str

    def __init__(self, code: int, message: str) -> None:
        super().__init__(message, code)
        self.code = code
        self.message = message

    def __str__(self) -> str:
        return f"AndesiteError ({self.code}): {self.message}"


class Searcher(Enum):
    """Supported search engines for Andesite."""
    YOUTUBE = "ytsearch"
    SOUNDCLOUD = "scsearch"


SearcherType = Union[Searcher, str]


def get_searcher(searcher: SearcherType) -> Searcher:
    """Get the `Searcher` for the given `SearcherType`.

    Args:
        searcher: Searcher to resolve. If searcher happens to be
            of type `Searcher` already, it is simply returned.

    This function can resolve the following `Searcher` specifications:

    - `Searcher` instance
    - Searcher id (i.e. "ytsearch", "scsearch")
    - Service name (i.e. "youtube", "soundcloud"). Note that the casing doesn't
      matter, as the provided names are converted to uppercase.

    Raises:
        TypeError: Invalid searcher type passed.
        ValueError: If searcher is a string but doesn't resolve to a valid
            searcher.
    """
    if isinstance(searcher, Searcher):
        return searcher
    elif isinstance(searcher, str):
        try:
            return Searcher[searcher.upper()]
        except KeyError:
            # if this causes a ValueError, just let it raise
            return Searcher(searcher)
    else:
        raise TypeError(f"Can only resolve {SearcherType}, not {type(searcher)}: {searcher}")


class AbstractHTTP(abc.ABC):
    """Abstract base class which requires a request method and a close method."""
    __slots__ = ()

    @property
    @abc.abstractmethod
    def closed(self) -> bool:
        """Whether or not the client is closed.

        If the client is closed it is no longer usable.
        """
        ...

    @abc.abstractmethod
    async def close(self) -> None:
        """Close the underlying connections and clean up.

        This should be called when you no longer need the client.
        """
        ...

    @abc.abstractmethod
    async def reset(self) -> None:
        """Reset the client so it may be used again.

        This has the opposite effect of the `close` method making the client
        usable again.
        """
        ...

    @abc.abstractmethod
    async def request(self, method: str, path: str, **kwargs) -> Any:
        """Perform a request and return the JSON response.

        Args:
            method: HTTP method to use
            path: Path relative to the base url. Must not start with a slash!
            **kwargs: Keyword arguments passed to the request

        This method is used by all other methods to perform their respective
        task. You should use the provided methods whenever possible.

        Raises:
            HTTPError: If Andesite returns an error.
        """
        ...


class HTTPInterface(AbstractHTTP, abc.ABC):
    """Abstract implementation of the endpoints.

    This does not include the player routes, as they are already covered by the `WebSocket`.
    The client uses the user agent `USER_AGENT` for every request.
    """
    __slots__ = ()

    async def get_stats(self) -> andesite.Stats:
        """Get the node's statistics.

        Raises:
            HTTPError: If Andesite returns an error.
        """
        data = await self.request("GET", "stats")
        return build_from_raw(andesite.Stats, data)

    async def load_tracks(self, identifier: str) -> andesite.LoadedTrack:
        """Load tracks.

        Args:
            identifier: Identifier to load. The identifier isn't handled in any way
                so it supports the search syntax for example.

        Raises:
            HTTPError: If Andesite returns an error.

        See Also:
            `search_tracks` to search for a track using a query.
        """
        data = await self.request("GET", "loadtracks", params=dict(identifier=identifier))
        return build_from_raw(andesite.LoadedTrack, data)

    async def load_tracks_safe(self, uri: str) -> andesite.LoadedTrack:
        """Load tracks from url.

        This is different from `load_tracks` insofar that it ignores
        special markers such as "ytsearch:" and treats the given uri
        as nothing but that.

        Args:
            uri: URI to load

        Raises:
            HTTPError: If Andesite returns an error.

        See Also:
            `load_tracks` to load a track using an identifier.
            `search_tracks` to search for a track using a query.
        """
        return await self.load_tracks(f"raw:{uri}")

    async def search_tracks(self, query: str, *,
                            searcher: SearcherType = Searcher.YOUTUBE) -> andesite.LoadedTrack:
        """Search tracks.

        Args:
            query: Search query to search for
            searcher: Specify the searcher to use. Defaults to YouTube.
                See `Searcher` for the supported searchers.

        Raises:
            HTTPError: If Andesite returns an error.

        Notes:
            This is a utility method for the `load_tracks` method.
            A search query is just an identifier with the format
            "<searcher>:<query>".
        """
        searcher_id = get_searcher(searcher).value
        return await self.load_tracks(f"{searcher_id}:{query}")

    async def decode_track(self, track: str) -> Optional[andesite.TrackInfo]:
        """Get the `TrackInfo` from the encoded track data.

        Notes:
            If you find yourself using this method a lot, you might want to use
            `lptrack <https://github.com/gieseladev/lptrack>`_ which can decode
            and encode the track data locally.

        Args:
            track: base 64 encoded track data to decode.

        Returns:
            `TrackInfo` of the provided data, `None` if the data is invalid.
            Note that this method doesn't raise `HTTPError`!
            If you need the `HTTPError` to be raised, use `decode_tracks`.

        See Also:
            Please use `decode_tracks` if you need to decode multiple
            encoded strings at once!
        """
        try:
            data = await self.request("POST", "decodetrack", json=dict(track=track))
        except HTTPError as e:
            if log.isEnabledFor(logging.DEBUG):
                log.debug(f"Couldn't decode track: {e}")
            return None
        else:
            return build_from_raw(andesite.TrackInfo, data)

    async def decode_tracks(self, tracks: Iterable[str]) -> List[andesite.TrackInfo]:
        """Get the `TrackInfo` from multiple encoded track data strings.

        Args:
            tracks: `Iterable` of base 64 encoded track data to decode.

        Returns:
            List of `TrackInfo` in order of the provided tracks.

        Raises:
            HTTPError: If Andesite returns an error.
        """
        data = await self.request("POST", "decodetracks", json=list(tracks))
        seq_build_all_items_from_raw(data, andesite.TrackInfo)
        return data


class HTTPBase(AbstractHTTP):
    """Standard implementation of `AbstractHTTP`.

    Args:
        password: Password to use for authorization. Use `None` if
            the Andesite node does not have a password set.

    See Also:
        `HTTP` for the client which includes the
            `HTTPInterface` methods.
    """

    __base_url: yarl.URL
    __session: Optional[aiohttp.ClientSession]
    __headers: Dict[str, str]

    def __init__(self, uri: Union[str, yarl.URL], password: Optional[str]) -> None:
        self.__base_url = yarl.URL(uri)

        headers = {"User-Agent": USER_AGENT}

        if password is not None:
            headers["Authorization"] = password

        self.__headers = headers

        self.__session = None

    def __repr__(self) -> str:
        return f"{type(self).__name__}(uri={self.__base_url!r}, password=[HIDDEN])"

    def __str__(self) -> str:
        return f"{type(self).__name__}({self.__base_url})"

    @property
    def aiohttp_session(self) -> aiohttp.ClientSession:
        """Client session used to make requests."""
        if self.__session is None:
            log.info("creating aiohttp client session for %s", self)
            self.__session = aiohttp.ClientSession(headers=self.__headers)

        return self.__session

    @property
    def closed(self) -> bool:
        if self.__session is not None:
            return self.__session.closed

        return False

    async def close(self) -> None:
        if self.__session:
            log.info("%s: closing aiohttp session", self)
            await self.__session.close()
        else:
            log.debug("%s: called close without ever creating a session, doing nothing", self)

    async def reset(self) -> None:
        log.debug("resetting ")
        self.__session = None

    async def request(self, method: str, path: str, **kwargs) -> Any:
        url = self.__base_url / path

        if log.isEnabledFor(logging.DEBUG):
            log.debug(f"performing {method} request for endpoint {path} with arguments: {kwargs}")

        async with self.aiohttp_session.request(method, url, **kwargs) as resp:
            data = await resp.json(content_type=None)

            if log.isEnabledFor(logging.DEBUG):
                log.debug(f"got data: {data}")

            if resp.status >= 400:
                try:
                    code = data["code"]
                    message = data["message"]
                except KeyError:
                    log.debug("Couldn't extract keys \"code\" and \"message\" from data, letting aiohttp raise!")
                    resp.raise_for_status()
                else:
                    raise HTTPError(code, message)

            return data


class HTTP(HTTPBase, HTTPInterface):
    """Client for Andesite's HTTP endpoints.

    See Also:
        `HTTPBase` for more details.
    """
    ...



"""Transformation utilities.

These functions are used to transform the data sent by Andesite
into the Python models.

These functions aren't exported to the `andesite` namespace, if you
want to use them you need to import them from `andesite.transform`.
However, there shouldn't be a need for you to use them in the first place
as andesite.py already does this for you.

Attributes:
    RawDataType (Dict[str, Any]): (Type alias) JSON-like object data type.
    CONVERTER_MEMO (lettercase.ConversionMemo): Memory used for letter case conversion.
    MapFunction ((T) -> `Any`): (Type alias) Callable that takes an argument and returns a new one.
"""

import dataclasses
from functools import partial
from typing import Any, Callable, Dict, MutableMapping, MutableSequence, Optional, Type, TypeVar, cast, overload

import lettercase

__all__ = ["RawDataType",
           "transform_input", "transform_output",
           "convert_to_raw", "build_from_raw",
           "seq_build_all_items_from_raw",
           "map_build_all_values_from_raw",
           "MapFunction", "map_convert_value", "map_convert_values", "map_convert_values_all",
           "map_build_values_from_raw",
           "from_milli", "to_milli",
           "from_centi", "to_centi",
           "map_convert_values_from_milli", "map_convert_values_to_milli",
           "map_filter_none", "map_rename_keys", "map_remove_keys"]

T = TypeVar("T")
KT = TypeVar("KT")
VT = TypeVar("VT")

RawDataType = Dict[str, Any]

# memo used to speed up key conversion
CONVERTER_MEMO = lettercase.ConversionMemo()


def _transform(transformer: Callable[[RawDataType], Optional[RawDataType]], data: RawDataType) -> RawDataType:
    """Perform a transformation.

    Args:
        transformer: Transformer which is called with the data
        data: Data to be transformed

    Returns:
        If the transformer returned something other than `None`, it is returned.
        Otherwise it is assumed, that the transformer manipulated the data
        and the original data is returned.
    """
    res = transformer(data)
    if res is None:
        return data
    else:
        return res


def transform_input(cls: Any, data: RawDataType) -> RawDataType:
    """Call the __transform_input__ classmethod on a model.

    This is different from calling the method directly because it
    always returns the current data.

    Args:
        cls: Target model whose transformation to apply
        data: Data to be transformed

    The transform method can either manipulate the provided data
    or replace it entirely by returning something other than `None`.

    Notes:
        When using this function inside of a transformation, make sure
        that you continue to work on the data returned by this function.
        Also, you cannot return `None` when using this function, you have
        to return the data. This is because you can't be sure whether the
        data has been modified or replaced and modifications on a replaced
        value wouldn't propagate upward unless you return them.
    """
    try:
        transformer = cls.__transform_input__
    except AttributeError:
        return data
    else:
        return _transform(transformer, data)


def transform_output(cls: Any, data: RawDataType) -> RawDataType:
    """Call the __transform_output__ classmethod on a model.

    This is different from calling the method directly because it
    always returns the current data.

    Args:
        cls: Target model whose transformation to apply
        data: Data to be transformed

    The transform method can either manipulate the provided data
    or replace it entirely by returning something other than `None`.

    Notes:
        When using this function inside of a transformation, make sure
        that you continue to work on the data returned by this function.
        Also, you cannot return `None` when using this function, you have
        to return the data. This is because you can't be sure whether the
        data has been modified or replaced and modifications on a replaced
        value wouldn't propagate upward unless you return them.
    """
    try:
        transformer = cls.__transform_output__
    except AttributeError:
        return data
    else:
        return _transform(transformer, data)


def convert_to_raw(obj: Any) -> RawDataType:
    """Convert a dataclass to a `dict`.

    This behaves similar to `dataclasses.asdict`. The difference is outlined below.

    The function uses `transform_output`, if the model provides a `__transform_output__` method
    it will be called.

    After transformation the keys of the data dict are converted from
    snake_case to dromedaryCase.

    This does not copy the values of the dataclass, modifying a value
    of the resulting `dict` will also modify the dataclass' value.

    Args:
        obj: Object to convert to its raw representation.
            Usually this is a dataclass, however, you can also parse
            `list`, `tuple`, and `dict` objects which will convert its
            members. All other types will be returned without modifying them.
    """
    if dataclasses.is_dataclass(obj):
        data: RawDataType = {}

        for field in dataclasses.fields(obj):
            field = cast(dataclasses.Field, field)
            value = convert_to_raw(getattr(obj, field.name))
            data[field.name] = value

        data = transform_output(obj, data)

        lettercase.mut_convert_keys(data, lettercase.SNAKE_CASE, lettercase.DROMEDARY_CASE, memo=CONVERTER_MEMO)

        return data
    elif isinstance(obj, (list, tuple)):
        return type(obj)(convert_to_raw(value) for value in obj)
    elif isinstance(obj, dict):
        return type(obj)((convert_to_raw(key), convert_to_raw(value)) for key, value in obj.items())
    else:
        # could copy it here to create a "safely" mutable dict but nah.
        return obj


@overload
def build_from_raw(cls: Type[T], raw_data: None) -> None: ...


@overload
def build_from_raw(cls: Type[T], raw_data: RawDataType) -> T: ...


def build_from_raw(cls: Type[T], raw_data: Optional[RawDataType]) -> Optional[T]:
    """Build an instance of cls from the passed raw data.

    In the spirit of the other transform functions, `None` is treated as a special value
    and returned directly instead of handling it in any way.


    The function uses `transform_input`, if the model provides a `__transform_input__` method
    it will be called.

    After transformation the data is used as the keyword arguments to the cls constructor.

    Args:
        cls: Target type to build
        raw_data: Data which should be used to build the instance.
            If this is `None`, `None` is returned.
    """
    if raw_data is None:
        return None

    lettercase.mut_convert_keys(raw_data, lettercase.DROMEDARY_CASE, lettercase.SNAKE_CASE, memo=CONVERTER_MEMO)
    raw_data = transform_input(cls, raw_data)

    return cls(**raw_data)


def seq_build_all_items_from_raw(items: MutableSequence[RawDataType], cls: Type[T]) -> None:
    """Build all items of a mutable sequence.

    This calls `build_from_raw` on all items in the sequence and assigns
    the result to the index.

    Args:
        items: Mutable sequence of raw data to be converted
        cls: Target type to build.

    Returns:
        This method mutates the provided sequence, it does not return anything.
    """
    for i, value in enumerate(items):
        items[i] = build_from_raw(cls, value)


MapFunction = Callable[[T], Any]


def map_convert_value(mapping: MutableMapping[KT, T], key: KT, func: MapFunction) -> None:
    """Call a function on the value of a key of the provided mapping.

    The return value of the function then replaces the old value.

    If the key does not exist in the mapping, it is ignored and the
    function is not called.

    Args:
        mapping: Mutable mapping which is to be manipulated.
        key: Key whose value in the mapping is to be converted.
            If this key does not exist in the mapping, the conversion
            is aborted and the function doesn't perform any action.
        func: Callback which will be called with the value of the key.
            Its return value then replaces the previous value in the
            mapping.

    Returns:
        This method mutates the provided mapping, it does not return anything.
    """
    try:
        value = mapping[key]
    except KeyError:
        return

    mapping[key] = func(value)


def map_convert_values(mapping: RawDataType, **key_funcs: MapFunction) -> None:
    """Run a callback for a key.

    For each key you can specify which function to run (<key> = <MapFunction>).

    Args:
        mapping: Mutable mapping for which to apply the conversion
        **key_funcs: key -> map function mapping.
            For each key the specified function will
            be applied using `map_convert_value`.

    Returns:
        This method mutates the provided mapping, it does not return anything.
    """
    for key, func in key_funcs.items():
        map_convert_value(mapping, key, func)


def map_convert_values_all(mapping: RawDataType, func: MapFunction, *keys: str) -> None:
    """Run the same callback on all keys.

    Works like `map_convert_values` but runs the same function for all keys.

    Args:
        mapping: Mutable mapping for which to apply the conversion
        func: Function to apply to the values of the specified keys.
            The function is run using `map_convert_value`
        *keys: Keys whose values are to be converted.

    Returns:
        This method mutates the provided mapping, it does not return anything.
    """
    for key in keys:
        map_convert_value(mapping, key, func)


def map_build_values_from_raw(mapping: RawDataType, **key_types: Type[T]) -> None:
    """Build the values of the specified keys to the specified type.

    Args:
        mapping: Mutable mapping for which to apply the conversion
        **key_types: key -> type mapping.
            For each key the value will be converted to the provided type
            using `build_from_raw`.

    Returns:
        This method mutates the provided mapping, it does not return anything.
    """
    for key, cls in key_types.items():
        func = partial(build_from_raw, cls)
        map_convert_value(mapping, key, func)


def map_build_all_values_from_raw(mapping: MutableMapping[Any, RawDataType], cls: Type[T]) -> None:
    """Build all values of a mapping.

    This calls `build_from_raw` on all values of the mapping
    and replaces the old value with the result.

    Args:
        mapping: Mutable mapping whose values are to be built
        cls: Type to convert the values to

    Returns:
        This method mutates the provided mapping, it does not return anything.
    """
    for key, value in mapping.items():
        mapping[key] = build_from_raw(cls, value)


@overload
def from_milli(value: int) -> float: ...


@overload
def from_milli(value: None) -> None: ...


def from_milli(value: Optional[int]) -> Optional[float]:
    """Convert a number from thousandths to base.

    Args:
        value: Value to convert from milli.

    Returns:
        Optional[float]: `None` if you pass `None` as the value, otherwise a `float`.
    """
    if value is None:
        return value

    return value / 1000


@overload
def to_milli(value: float) -> int: ...


@overload
def to_milli(value: None) -> None: ...


def to_milli(value: Optional[float]) -> Optional[int]:
    """Convert from base unit to thousandths.

    Args:
        value: Value to convert to milli.

    Returns:
        Optional[int]: `None` if you pass `None` as the value, otherwise an `int`.
    """
    if value is None:
        return value

    return round(1000 * value)


@overload
def from_centi(value: int) -> float: ...


@overload
def from_centi(value: None) -> None: ...


def from_centi(value: Optional[int]) -> Optional[float]:
    """Convert a number from hundredths to base.

    Args:
        value: Value to convert from centi.

    Returns:
        Optional[float]: `None` if you pass `None` as the value, otherwise a `float`.
    """
    if value is None:
        return value

    return value / 100


@overload
def to_centi(value: float) -> int: ...


@overload
def to_centi(value: None) -> None: ...


def to_centi(value: Optional[float]) -> Optional[int]:
    """Convert from base unit to hundredths.

    This is really just multiplying by 1000.

    Args:
        value: Value to convert to milli.

    Returns:
        Optional[int]: `None` if you pass `None` as the value, otherwise an `int`.
    """
    if value is None:
        return value

    return round(100 * value)


def map_convert_values_from_milli(mapping: RawDataType, *keys) -> None:
    """Run `from_milli` on all specified keys' values.

    Args:
        mapping: Mutable mapping for which to apply the conversion
        *keys: Keys whose values to convert from milli.
            Uses `from_milli` to perform the conversion.

    Returns:
        This method mutates the provided mapping, it does not return anything.
    """
    map_convert_values_all(mapping, from_milli, *keys)


def map_convert_values_to_milli(mapping: RawDataType, *keys) -> None:
    """Run `to_milli` on all specified keys' values.

    Args:
        mapping: Mutable mapping for which to apply the conversion
        *keys: Keys whose values to convert to milli.
            Uses `to_milli` to perform the conversion.

    Returns:
        This method mutates the provided mapping, it does not return anything.
    """
    map_convert_values_all(mapping, to_milli, *keys)


def map_filter_none(mapping: MutableMapping[Any, Any]) -> None:
    """Remove all keys from the mapping whose values are `None`.

    Args:
        mapping: Mutable mapping to filter

    Returns:
        This method mutates the provided mapping, it does not return anything.
    """
    remove_keys = {key for key, value in mapping.items() if value is None}

    for key in remove_keys:
        del mapping[key]


def map_rename_keys(mapping: MutableMapping[str, Any], **key_maps: str) -> None:
    """Rename keys of a mapping.
    This is just deleting the old key and assigning its value to the new key.

    Args:
        mapping: Mutable mapping to manipulate
        **key_maps: new -> old name mapping.
            The reason the mapping isn't from old to new is that you may want
            to rename keys that aren't python-friendly.
    """

    for new_key, old_key in key_maps.items():
        try:
            value = mapping.pop(old_key)
        except KeyError:
            continue
        else:
            mapping[new_key] = value


def map_remove_keys(mapping: MutableMapping[KT, Any], *keys: KT) -> None:
    """Remove a number of keys (and their values) from a mapping.

    Args:
        mapping: Mapping to remove keys from.
        *keys: Keys to remove.

    Returns:
        Nothing, the operation mutates the given mapping.
    """
    for key in keys:
        try:
            del mapping[key]
        except KeyError:
            pass



"""Sendable operations to Andesite.

These operations can be sent using `AbstractWebSocket.send_operation`.

See Also:
    `andesite.models.receive_operations` for operations sent by Andesite.

Attributes:
    MixerPlayerUpdateMap (Dict[str, Union[Play, Update]]): (Type alias) str -> `Play`/`Update` map used by the `MixerUpdate`.
"""

import abc
from dataclasses import dataclass
from typing import Any, Dict, Optional, Union

from andesite.transform import RawDataType, build_from_raw, from_centi, from_milli, map_build_values_from_raw, \
    map_convert_values, map_convert_values_from_milli, map_convert_values_to_milli, to_centi, to_milli
from .filters import FilterMap, FilterMapLike

__all__ = ["SendOperation",
           "VoiceServerUpdate",
           "Play", "Pause", "Seek", "Volume", "FilterUpdate", "Update", "MixerUpdate"]


class SendOperation(abc.ABC):
    """SendOperation is a model that can be passed as a payload to the
    `WebSocket` client.

    See Also:
        `WebSocket.send_operation`
    """
    __op__: str


@dataclass
class VoiceServerUpdate(SendOperation):
    """Operation providing a voice server update.

    Attributes:
        session_id (str): Session ID for the current user in the event's guild
        event (Dict[str, Any]): Voice server update event sent by discord
    """
    __op__ = "voice-server-update"

    session_id: str
    event: Dict[str, Any]


@dataclass
class Play(SendOperation):
    """Operation playing a track.

    Attributes:
        track (str): base64 encoded lavaplayer track
        start (Optional[float]): timestamp, in seconds, to start the track
        end (Optional[float]): timestamp, in seconds, to end the track
        pause (Optional[bool]): whether or not to pause the player
        volume (Optional[float]): volume to set on the player
        no_replace (bool): if `True` and a track is already playing/paused, this command is ignored. (Defaults to `False`)
    """
    __op__ = "play"

    track: str
    start: Optional[float] = None
    end: Optional[float] = None
    pause: Optional[bool] = None
    volume: Optional[float] = None
    no_replace: bool = False

    @classmethod
    def __transform_input__(cls, data: RawDataType) -> None:
        map_convert_values_from_milli(data, "start", "end")
        map_convert_values(data, volume=from_centi)

    @classmethod
    def __transform_output__(cls, data: RawDataType) -> None:
        map_convert_values_to_milli(data, "start", "end")
        map_convert_values(data, volume=to_centi)


@dataclass
class Pause(SendOperation):
    """Operation pausing the player.

    Attributes:
        pause (bool): whether or not to pause the player
    """
    __op__ = "pause"

    pause: bool


@dataclass
class Seek(SendOperation):
    """Operation seeking the current track.

    Attributes:
        position (float): timestamp to set the current track to, in seconds
    """
    __op__ = "seek"

    position: float

    @classmethod
    def __transform_input__(cls, data: RawDataType) -> None:
        map_convert_values_from_milli(data, "position")

    @classmethod
    def __transform_output__(cls, data: RawDataType) -> None:
        map_convert_values_to_milli(data, "position")


@dataclass
class Volume(SendOperation):
    """Operation adjusting the volume.

    Attributes:
        volume (float): volume to set on the player
    """
    __op__ = "volume"

    volume: float

    @classmethod
    def __transform_input__(cls, data: RawDataType) -> None:
        map_convert_values(data, volume=from_centi)

    @classmethod
    def __transform_output__(cls, data: RawDataType) -> None:
        map_convert_values(data, volume=to_centi)


@dataclass
class FilterUpdate(FilterMap, SendOperation):
    """Operation adjusting the filter settings.

    See Also:
        This class inherits from `FilterMap`.
    """
    __op__ = "filters"

    def __init__(self, filters: FilterMapLike) -> None:
        super().__init__(filters)


@dataclass
class Update(SendOperation):
    """Operation providing an update for the current track.

    Attributes:
        pause (Optional[bool]): whether or not to pause the player
        position (Optional[float]): timestamp to set the current track to, in seconds
        volume (Optional[float]): volume to set on the player
        filters (Optional[FilterUpdate]): configuration for the filters
    """
    __op__ = "update"

    pause: Optional[bool] = None
    position: Optional[float] = None
    volume: Optional[float] = None
    filters: Optional[FilterUpdate] = None

    @classmethod
    def __transform_input__(cls, data: RawDataType) -> None:
        map_convert_values(data, volume=from_centi, position=from_milli)
        map_build_values_from_raw(data, filters=FilterUpdate)

    @classmethod
    def __transform_output__(cls, data: RawDataType) -> None:
        map_convert_values(data, volume=to_centi, position=to_milli)


MixerPlayerUpdateMap = Dict[str, Union[Play, Update]]


@dataclass
class MixerUpdate(SendOperation):
    """Operation adjusting the mixer players.

    Attributes:
        enable (Optional[bool]): if present, controls whether or not the mixer should be used
        players (MixerPlayerUpdateMap): map of player id to `Play` / `Update` payloads for each mixer source
    """
    __op__ = "mixer"

    enable: Optional[bool]
    players: MixerPlayerUpdateMap

    @classmethod
    def __transform_input__(cls, data: RawDataType) -> None:
        players = data["players"]
        for key, value in players.items():
            try:
                new_value = build_from_raw(Play, value)
            except TypeError:
                new_value = build_from_raw(Update, value)

            players[key] = new_value



"""Event models for the web socket client."""

from __future__ import annotations

import dataclasses
import logging
from typing import Any, Dict, TypeVar

import andesite

__all__ = ["WebSocketConnectEvent", "WebSocketDisconnectEvent",
           "RawMsgReceiveEvent",
           "RawMsgSendEvent"]

ROPT = TypeVar("ROPT", bound=andesite.ReceiveOperation)
ET = TypeVar("ET", bound=andesite.AndesiteEvent)

log = logging.getLogger(__name__)


@dataclasses.dataclass()
class WebSocketConnectEvent:
    """Event dispatched when a connection has been established.

    Attributes:
        client (AbstractWebSocketClient): Web socket client which
            connected.
    """
    client: andesite.AbstractWebSocketClient


@dataclasses.dataclass()
class WebSocketDisconnectEvent:
    """Event dispatched when a client was disconnected.

    Attributes:
        client (AbstractWebSocketClient): Web socket client which connected.
        deliberate (bool): Whether the disconnect was deliberate.
    """

    client: andesite.AbstractWebSocketClient
    deliberate: bool


@dataclasses.dataclass()
class RawMsgReceiveEvent:
    """Event emitted when a web socket message is received.

    Attributes:
        client (AbstractWebSocket): Web socket client that
            received the message.
        body (Dict[str, Any]): Raw body of the received message.
            Note: The body isn't manipulated in any way other
            than being loaded from the raw JSON string.
            For example, the names are still in dromedaryCase.
    """
    client: andesite.AbstractWebSocketClient
    body: Dict[str, Any]


@dataclasses.dataclass()
class RawMsgSendEvent:
    """Event dispatched before a web socket message is sent.

    It's important to note that this is not a receipt of a message being sent,
    this event is dispatched even if the message fails to send.

    Attributes:
        client (AbstractWebSocketClient): Web socket client that
            received the message.
        guild_id (int): guild id
        op (str): Op-code to be executed
        body (Dict[str, Any]): Raw body of the message
    """
    client: andesite.AbstractWebSocketClient
    guild_id: int
    op: str
    body: Dict[str, Any]



"""Web socket client for Andesite.

Use `WebSocket` if you just want a client which
connects to a single Andesite node.
"""

from __future__ import annotations

import abc
import asyncio
import logging
import math
import time
from collections import deque
from contextlib import suppress
from json import JSONDecodeError, JSONDecoder, JSONEncoder
from typing import Any, Deque, Dict, Optional, Tuple, Type, TypeVar, Union, overload

import aiobservable
import websockets
from websockets import ConnectionClosed, InvalidHandshake, WebSocketClientProtocol
from websockets.http import Headers
from yarl import URL

import andesite
from .transform import build_from_raw, convert_to_raw, map_filter_none, to_centi, to_milli
from .web_socket_client_events import RawMsgReceiveEvent, RawMsgSendEvent, WebSocketConnectEvent, \
    WebSocketDisconnectEvent

__all__ = ["try_connect",
           "AbstractWebSocket", "AbstractWebSocketClient",
           "WebSocketInterface",
           "WebSocketBase",
           "WebSocket"]

ROPT = TypeVar("ROPT", bound=andesite.ReceiveOperation)
ET = TypeVar("ET", bound=andesite.AndesiteEvent)

log = logging.getLogger(__name__)


async def try_connect(uri: str, **kwargs) -> Optional[WebSocketClientProtocol]:
    """Connect to the given uri and return the client.

    Catches exceptions which could potentially be solved
    by retrying.

    Args:
        uri: URI to connect to
        **kwargs: keyword arguments to pass to the `websockets.connect` function.

    Returns:
        `WebSocketClientProtocol` if the connection succeeded, None otherwise.
    """
    try:
        client = await websockets.connect(uri, **kwargs)
    except (OSError, InvalidHandshake) as e:
        log.info(f"Connection to {uri} failed: {e}")
        return None

    return client


def _get_ops_for_player(track: str, player: Optional[andesite.BasePlayer]) -> Tuple[andesite.Play, andesite.Update]:
    """Get operations which apply the state from the given player.

    Two operations are required to both play the track and set the filters.

    Args:
        track: Track to play
        player: Player to get state from.

    Returns:
        Two operations which when sent to Andesite will apply the given player's
        state.
    """
    if player is not None:
        play_op = andesite.Play(track,
                                start=player.position,
                                pause=player.paused,
                                volume=player.volume,
                                no_replace=False)

        update_op = andesite.Update(
            pause=player.paused,
            position=player.position,
            volume=player.volume,
            filters=andesite.FilterUpdate(player.filters),
        )
    else:
        play_op = andesite.Play(track)
        update_op = andesite.Update()

    return play_op, update_op


class AbstractWebSocket(abc.ABC):
    """Abstract base class for an Andesite web socket client.

    This class is separate from `AbstractWebSocketClient` to
    support more complex clients which use more than one web socket
    connection (ex: Pools).

    See Also:
        `AbstractWebSocketClient` for the abstract base class
        for actual web socket clients.


    :Events:
        - **ws_connect** (`WebSocketConnectEvent`): When the client connects
        - **ws_disconnect** (`WebSocketDisconnectEvent`): When the client
          disconnects

        - **raw_msg_receive** (`RawMsgReceiveEvent`): Whenever a message body is
          received. For this event to be dispatched, the received message needs
          to be valid JSON and an object.
        - **raw_msg_send** (`RawMsgSendEvent`): When sending a message.
          This event is dispatched regardless of whether the message
          was actually sent.

        - **player_update** (`PlayerUpdate`): When a `PlayerUpdate` is received.

        - **track_start** (`TrackStartEvent`)
        - **track_end** (`TrackEndEvent`)
        - **track_exception** (`TrackExceptionEvent`)
        - **track_stuck** (`TrackStuckEvent`)
        - **unknown_andesite_event** (`UnknownAndesiteEvent`): When an unknown
          event is received.
    """
    __event_target: aiobservable.Observable
    __state_handler: Optional[andesite.AbstractState]

    @property
    def event_target(self) -> aiobservable.Observable:
        """Event target to send events to.

        If no event target is set, but the instance is itself an event target,
        self is set as the new event target and returned, otherwise a new event
        target is created.
        """
        try:
            return self.__event_target
        except AttributeError:
            if isinstance(self, aiobservable.Observable):
                self.__event_target = self
            else:
                self.__event_target = aiobservable.Observable()

            return self.__event_target

    @property
    def state(self) -> Optional[andesite.AbstractState]:
        """State handler for the client.

        You may manually set this value to a different state handler which
        implements the `AbstractState`. You can also disable state handling
        by setting it to `False`.
        Note that this won't apply the state to the client! You can use the
        `load_player_state` method do load individual player states.

        If not state is set the getter either returns the current instance, if
        it happens to implement `AbstractState`, otherwise it returns
        `None`.
        """
        try:
            return self.__state_handler
        except AttributeError:
            if isinstance(self, andesite.AbstractState):
                self.__state_handler = self
            else:
                self.__state_handler = None

            return self.__state_handler

    @state.setter
    def state(self, state: andesite.StateArgumentType) -> None:
        self.__state_handler = andesite._get_state(state)

    @property
    @abc.abstractmethod
    def closed(self) -> bool:
        """Whether or not the client is closed.

        If the client is closed it is no longer usable.
        """
        ...

    @abc.abstractmethod
    async def close(self) -> None:
        """Close the underlying connections and clean up.

        This should be called when you no longer need the client.
        """
        ...

    @abc.abstractmethod
    async def reset(self) -> None:
        """Reset the client so it may be used again.

        This has the opposite effect of the `close` method making the client
        usable again.
        """
        ...

    @abc.abstractmethod
    async def send(self, guild_id: int, op: str, payload: Dict[str, Any]) -> None:
        """Send a payload.

        The guild_id and op are written to the payload before it is converted to JSON.

        If sending the message fails because the connection is closed,
        it is added to the message queue and a connection attempt is started.

        Regardless of whether the payload was actually sent or
        not a `RawMsgSendEvent` (`raw_msg_send`) event is dispatched.

        Args:
            guild_id: Target guild id
            op: Name of operation to perform
            payload: Additional data to be sent along with the data.
                The payload is mutated by the function.
        """
        ...

    async def send_operation(self, guild_id: int, operation: andesite.SendOperation) -> None:
        """Send a `SendOperation`.

        Args:
            guild_id: Target guild id
            operation: Operation to send

        Notes:
            Using `SendOperation` instances to send messages is slightly
            less efficient than calling the respective `WebSocket` methods
            directly.
        """
        await self.send(guild_id, operation.__op__, convert_to_raw(operation))

    async def load_player_state(self, player_state: andesite.AbstractPlayerState) -> None:
        """Load a player state.

        Args:
            player_state: State to load.
        """
        guild_id = player_state.guild_id

        last_update = await player_state.get_voice_server_update()
        if last_update:
            await self.send_operation(guild_id, last_update)

        # TODO can we do mixer players?

        track, player = await asyncio.gather(
            player_state.get_track(),
            player_state.get_player(),
        )

        if track:
            if last_update is None:
                log.warning(f"loading player state without voice server update! ({self})")

            # of course because Andesite is... great [citation needed] we need
            # to use two operations to set it up properly...
            play_op, update_op = _get_ops_for_player(track, player)

            await asyncio.gather(
                self.send_operation(guild_id, play_op),
                self.send_operation(guild_id, update_op),
            )


class AbstractWebSocketClient(AbstractWebSocket, abc.ABC):
    """Abstract base class for a singular web socket connection to Andesite.

    If you're creating a new client and it doesn't use only one Andesite node,
    you should implement `AbstractWebSocket` instead.

    See Also:
        `AbstractWebSocket` for more details.
    """
    __slots__ = ()

    @property
    @abc.abstractmethod
    def connected(self) -> bool:
        """Whether the client is connected and usable."""
        ...

    @property
    @abc.abstractmethod
    def connection_id(self) -> Optional[str]:
        """Andesite connection id.

        This should be set after connecting to the node.
        The connection id can be used to resume connections.

        This property is deletable. The consequence of which is that it won't
        be sent the next time the client connects to a node.

        Notes:
            The client already performs resuming automatically.
        """
        ...

    @property
    @abc.abstractmethod
    def node_region(self) -> Optional[str]:
        """Node region sent by the Andesite node.

        This will be set after the connection is established.
        """
        ...

    @property
    @abc.abstractmethod
    def node_id(self) -> Optional[str]:
        """Node id sent by the Andesite node.

        This will be set after the connection is established.
        """
        ...

    @abc.abstractmethod
    async def connect(self, *, max_attempts: int = None) -> None:
        """Connect to the web socket server.

        Args:
            max_attempts: Amount of connection attempts to perform before aborting.
                If `None`, unlimited attempts will be performed.

        If `max_attempts` is exceeded and the client gives up on connecting it
        is closed!

        After successfully connecting all messages in the message
        queue are sent in the order they were added and a `ws_connected`
        event is dispatched.

        Notes:
            This method doesn't have to be called manually,
            it is called as soon as the first message needs to be sent.
            However, there are good reasons to call it anyway, such
            as the ability to check the validity of the URI, or to
            receive events from Andesite.
        """
        ...

    @abc.abstractmethod
    async def disconnect(self) -> None:
        """Disconnect the web socket.

        This will also stop the client from
        receiving events from Andesite.

        This method is idempotent, it doesn't do
        anything if the client isn't connected.

        Notes:
            This is different from `close`. Calling `close`
            disconnects the client and causes it to become unusable.
            This method only disconnects the client so it can be reconnected.
        """
        ...


class WebSocketInterface(AbstractWebSocket, abc.ABC):
    """Implementation of the web socket endpoints."""
    __slots__ = ()

    async def wait_for_receive(self, op_type: Type[ROPT], guild_id: int = None) -> ROPT:
        if guild_id is None:
            predicate = None
        else:
            def predicate(op: andesite.ReceiveOperation) -> bool:
                try:
                    # noinspection PyUnresolvedReferences
                    gid = op.guild_id
                except AttributeError:
                    return False

                return gid == guild_id

        return await self.event_target.subscribe(op_type).first(predicate=predicate)

    # noinspection PyOverloads
    @overload
    async def play(self, guild_id: int, track: andesite.Play) -> None:
        ...

    @overload
    async def play(self, guild_id: int, track: str, *,
                   start: float = None,
                   end: float = None,
                   pause: bool = None,
                   volume: float = None,
                   no_replace: bool = False) -> None:
        ...

    async def play(self, guild_id: int, track: Union[str, andesite.Play], *,
                   start: float = None,
                   end: float = None,
                   pause: bool = None,
                   volume: float = None,
                   no_replace: bool = False) -> None:
        """Play a track on the guild.

        Instead of providing all the fields you can also pass a `Play` operation.

        Args:
            guild_id: ID of the guild for which to play
            track: Either a `Play` operation or a base64 encoded lavaplayer track.
                If you pass a `Play` operation you must not set any of the
                keyword arguments.

            start: timestamp, in seconds, to start the track
            end: timestamp, in seconds, to end the track
            pause: whether or not to pause the player
            volume: volume to set on the player
            no_replace: if True and a track is already playing/paused,
                this command is ignored (Default: False)
        """
        if isinstance(track, andesite.Play):
            payload = convert_to_raw(track)
        else:
            payload = dict(track=track, start=to_milli(start), end=to_milli(end), pause=pause, volume=to_centi(volume),
                           noReplace=no_replace)
            map_filter_none(payload)

        await self.send(guild_id, "play", payload)

    async def pause(self, guild_id: int, pause: bool) -> None:
        """Pause a player.

        If the player is already paused or unpaused,
        this is a no-op.

        Args:
            guild_id: ID of the guild for which to pause
            pause: `True` to pause, `False` to unpause
        """
        await self.send(guild_id, "pause", dict(pause=pause))

    async def stop(self, guild_id: int) -> None:
        """Stop a player.

        Args:
            guild_id: ID of the guild for which to stop
        """
        await self.send(guild_id, "stop", {})

    async def seek(self, guild_id: int, position: float) -> None:
        """Seek to a position.

        Args:
            guild_id: ID of the guild for which to seek
            position: Timestamp, in seconds, to seek to
        """
        payload = dict(position=to_milli(position))
        await self.send(guild_id, "seek", payload)

    async def volume(self, guild_id: int, volume: float) -> None:
        """Set a player's volume.

        Args:
            guild_id: ID of the guild for which to set the volume
            volume: Volume to set
        """
        payload = dict(volume=to_centi(volume))
        await self.send(guild_id, "volume", payload)

    @overload
    async def update(self, guild_id: int, update: andesite.Update) -> None:
        ...

    @overload
    async def update(self, guild_id: int, *,
                     pause: bool = None,
                     position: float = None,
                     volume: float = None,
                     filters: andesite.FilterMapLike = None) -> None:
        ...

    async def update(self, guild_id: int, update: andesite.Update = None, *,
                     pause: bool = None,
                     position: float = None,
                     volume: float = None,
                     filters: andesite.FilterMapLike = None) -> None:
        """Send an update.

        You may either provide the given keyword arguments,
        or pass an `Update` operation which will be used instead.

        Args:
            guild_id: ID of the guild for which to update.
            update: You can provide an `Update` operation instead of
                the keyword arguments.

            pause: whether or not to pause the player
            position: timestamp to set the current track to, in seconds
            volume: volume to set on the player
            filters: configuration for the filters
        """
        if isinstance(update, andesite.Update):
            payload = convert_to_raw(update)
        else:
            if filters:
                filters = convert_to_raw(filters)

            payload = dict(pause=pause, position=to_milli(position), volume=to_centi(volume), filters=filters)
            map_filter_none(payload)

        await self.send(guild_id, "update", payload)

    async def destroy(self, guild_id: int) -> None:
        """Destroy a player.

        Args:
            guild_id: ID of the guild for which to destroy the player
        """
        await self.send(guild_id, "destroy", {})

    async def mixer(self, guild_id: int, enable: bool = None, **players: Union[andesite.Play, andesite.Update]) -> None:
        """Configure the mixer player.

        Args:
            guild_id: ID of the guild for which to configure the mixer
            enable: If present, controls whether or not the mixer should be used
            **players: Map of player id to `Play` / `Update` payloads for each mixer source
        """
        payload: Dict[str, Any] = convert_to_raw(players)
        payload["enable"] = enable

        await self.send(guild_id, "mixer", payload)

    @overload
    async def filters(self, guild_id: int, filter_update: andesite.FilterMap) -> None:
        ...

    @overload
    async def filters(self, guild_id: int, *,
                      equalizer: andesite.Equalizer = None,
                      karaoke: andesite.Karaoke = None,
                      timescale: andesite.Timescale = None,
                      tremolo: andesite.Tremolo = None,
                      vibrato: andesite.Vibrato = None,
                      volume: andesite.VolumeFilter = None) -> None:
        ...

    async def filters(self, guild_id: int, filter_update: andesite.FilterMap = None, *,
                      equalizer: andesite.Equalizer = None,
                      karaoke: andesite.Karaoke = None,
                      timescale: andesite.Timescale = None,
                      tremolo: andesite.Tremolo = None,
                      vibrato: andesite.Vibrato = None,
                      volume: andesite.VolumeFilter = None,
                      **custom_filters: Any) -> None:
        """Configure the filters of a player.

        Args:
            guild_id: ID of the guild for which to configure the filters
            filter_update: Instead of specifying the other keyword arguments
                you may provide a `FilterMap` operation which will be used
                instead.
            equalizer: Equalizer filter settings
            karaoke: Karaoke filter settings
            timescale: Timescale filter settings
            tremolo: Tremolo filter settings
            vibrato: Vibrato filter settings
            volume: Volume filter settings
            **custom_filters: Ability to specify additional filters
                that aren't supported by the library.
        """
        if isinstance(filter_update, andesite.FilterUpdate):
            payload = convert_to_raw(filter_update)
        else:
            payload = convert_to_raw(dict(
                equalizer=equalizer,
                karaoke=karaoke,
                timescale=timescale,
                tremolo=tremolo,
                vibrato=vibrato,
                volume=volume,
                **custom_filters
            ))

        await self.send(guild_id, "filters", payload)

    async def get_player(self, guild_id: int) -> andesite.Player:
        """Get the player.

        Args:
            guild_id: Target guild id

        Returns:
            Player for the guild
        """

        await self.send(guild_id, "get-player", {})
        player_update = await self.wait_for_receive(andesite.PlayerUpdate, guild_id=guild_id)
        return player_update.state

    async def get_stats(self, guild_id: int) -> andesite.Stats:
        """Get the Andesite stats.

        Args:
            guild_id: Target guild id

        Returns:
            Statistics for the node
        """
        await self.send(guild_id, "get-stats", {})
        stats_update = await self.wait_for_receive(andesite.StatsUpdate)
        return stats_update.stats

    async def ping(self, guild_id: int) -> float:
        """Ping the Andesite server.

        Args:
            guild_id: Target guild id

        Returns:
            Amount of seconds it took for the response to be received.
            Note: This is not necessarily an accurate reflection of the actual latency.
        """
        start = time.time()

        await asyncio.gather(
            self.send(guild_id, "ping", {}),
            self.wait_for_receive(andesite.PongResponse, guild_id=guild_id),
        )

        return time.time() - start

    async def voice_server_update(self, guild_id: int, session_id: str, event: Dict[str, Any]) -> None:
        """Provide a voice server update.

        Args:
            guild_id: ID of the guild of the voice server update
            session_id: session id
            event: voice server update event as sent by Discord

        Notes:
            If you wish to send a `VoiceServerUpdate` operation, please
            use the `send_operation` method directly.
        """
        payload = dict(sessionId=session_id, event=event)
        await self.send(guild_id, "voice-server-update", payload)


class WebSocketBase(AbstractWebSocketClient):
    """Client for the Andesite WebSocket handler.

    Args:
        ws_uri: Websocket endpoint to connect to.
        user_id: Bot's user id. If at the time of creation this is unknown,
            you may pass `None`, but then it needs to be set before connecting
            for the first time.
        password: Authorization for the Andesite node.
            Set to `None` if the node doesn't have a password.
        state: State handler to use. If `False` state handling is disabled.
            `None` to use the default state handler (`State`).
        max_connect_attempts: See the `max_connect_attempts` attribute.

    The client automatically keeps track of the current connection id and
    resumes the previous connection when calling `connect`, if there is any.
    You can delete the `connection_id` property to disable this.

    See Also:
        `AbstractWebSocketClient` for more details including a list
        of events that are dispatched.

    Attributes:
        max_connect_attempts (Optional[int]): Max amount of connection attempts
            to start before giving up. If `None`, there is no upper limit.
            This value can be overwritten when calling `connect` manually.
        web_socket_client (Optional[WebSocketClientProtocol]):
            Web socket client which is used.
            This attribute will be set once `connect` is called.
            Don't use the presence of this attribute to check whether
            the client is connected, use the `connected` property.
    """
    max_connect_attempts: Optional[int]

    web_socket_client: Optional[WebSocketClientProtocol]

    __closed: bool

    __ws_uri: str
    __headers: Headers
    __last_connection_id: Optional[str]

    __connect_lock: Optional[asyncio.Lock]
    __message_queue: Deque[str]

    __read_loop: Optional[asyncio.Future]

    _json_encoder: JSONEncoder
    _json_decoder: JSONDecoder

    def __init__(self, ws_uri: Union[str, URL], user_id: Optional[int], password: Optional[str], *,
                 state: andesite.StateArgumentType = False,
                 max_connect_attempts: int = None) -> None:
        self.__ws_uri = str(ws_uri)

        self.__headers = Headers()
        if password is not None:
            self.__headers["Authorization"] = password
        if user_id is not None:
            self.user_id = user_id

        self.__last_connection_id = None

        self.max_connect_attempts = max_connect_attempts

        self.web_socket_client = None

        # can't create the lock here, because if the user uses
        # asyncio.run and creates the client outside of it, the loop
        # within the lock will not be the same as the loop used by
        # asyncio.run (as it creates a new loop every time)
        self.__connect_lock = None
        self.__message_queue = deque()

        self.__closed = False
        self.__read_loop = None

        self._json_encoder = JSONEncoder()
        self._json_decoder = JSONDecoder()

        self.state = state

    def __repr__(self) -> str:
        return f"{type(self).__name__}(ws_uri={self.__ws_uri!r}, user_id={self.user_id!r}, " \
               f"password=[HIDDEN], state={self.state!r}, max_connect_attempts={self.max_connect_attempts!r})"

    def __str__(self) -> str:
        return f"{type(self).__name__}({self.__ws_uri})"

    @property
    def user_id(self) -> Optional[int]:
        """User id.

        This is only `None` if it wasn't passed to the constructor.

        You can set this property to a new user id.
        """
        return self.__headers.get("User-Id")

    @user_id.setter
    def user_id(self, user_id: int) -> None:
        self.__headers["User-Id"] = str(user_id)

    @property
    def closed(self) -> bool:
        return self.__closed

    @property
    def connected(self) -> bool:
        if self.web_socket_client:
            return self.web_socket_client.open
        else:
            return False

    @property
    def connection_id(self) -> Optional[str]:
        return self.__last_connection_id

    @connection_id.deleter
    def connection_id(self) -> None:
        self.__last_connection_id = None

    @property
    def node_region(self) -> Optional[str]:
        client = self.web_socket_client
        if client:
            return client.response_headers.get("Andesite-Node-Region")

        return None

    @property
    def node_id(self) -> Optional[str]:
        client = self.web_socket_client
        if client:
            return client.response_headers.get("Andesite-Node-Id")

        return None

    def _get_connect_lock(self, *, loop: asyncio.AbstractEventLoop = None) -> asyncio.Lock:
        """Get the connect lock.

        The connect lock is only created once. Subsequent calls always return
        the same lock. The reason for the delayed creating is that the lock is
        bound to an event loop, which can change between __init__ and connect.
        """
        if self.__connect_lock is None:
            self.__connect_lock = asyncio.Lock(loop=loop)

        return self.__connect_lock

    async def _connect(self, max_attempts: int = None) -> None:
        """Internal connect method.

        Args:
            max_attempts: Max amount of connection attempts to perform before aborting.
                This overwrites the instance attribute `max_connect_attempts`.

        Raises:
            ValueError: If client is already connected

        Notes:
            If `max_attempts` is exceeded and the client gives up on connecting it
            is closed!
        """
        if self.connected:
            raise ValueError("Already connected!")

        headers = self.__headers

        if "User-Id" not in headers:
            raise KeyError("Trying to connect but user id unknown.\n"
                           "This is most likely the case because you didn't\n"
                           "set the user_id in the constructor and forgot to\n"
                           "set it before connecting!")

        # inject the connection id to resume previous connection
        if self.__last_connection_id is not None:
            headers["Andesite-Resume-Id"] = self.__last_connection_id
        else:
            with suppress(KeyError):
                del headers["Andesite-Resume-Id"]

        attempt: int = 1
        max_attempts = max_attempts or self.max_connect_attempts

        while max_attempts is None or attempt <= max_attempts:
            client = await try_connect(self.__ws_uri, extra_headers=headers)
            if client:
                break

            timeout = int(math.pow(attempt, 1.5))
            log.info(f"Connection unsuccessful, trying again in {timeout} seconds")
            await asyncio.sleep(timeout)

            attempt += 1
        else:
            self.__closed = True
            raise ConnectionError(f"Couldn't connect to {self.__ws_uri} after {attempt} attempts")

        self.web_socket_client = client
        self._start_read_loop()

        _ = self.event_target.emit(WebSocketConnectEvent(self))

        await self._replay_message_queue()

    async def connect(self, *, max_attempts: int = None) -> None:
        if self.closed:
            raise ValueError("Client is closed and cannot be reused.")

        async with self._get_connect_lock():
            if not self.connected:
                await self._connect(max_attempts)

    async def disconnect(self) -> None:
        self._stop_read_loop()

        if self.connected:
            await self.web_socket_client.close(reason="disconnect")
            _ = self.event_target.emit(WebSocketDisconnectEvent(self, True))

    async def reset(self) -> None:
        await self.disconnect()
        self.__message_queue.clear()
        del self.connection_id

        self.__closed = False

    async def close(self) -> None:
        await self.disconnect()
        self.__closed = True

    async def _replay_message_queue(self) -> None:
        """Send all messages in the message queue."""
        if not self.__message_queue:
            return

        log.info(f"Sending {len(self.__message_queue)} queued messages")

        try:
            for msg in self.__message_queue:
                await self.web_socket_client.send(msg)
        finally:
            self.__message_queue.clear()

    async def _web_socket_reader(self) -> None:
        """Internal web socket read loop.

        This method should never be called manually, see the following
        methods for controlling the reader.

        See Also:
            `WebSocket._start_read_loop` to start the read loop.
            `WebSocket._stop_read_loop` to stop the read loop.

        Notes:
            The read loop is automatically managed by the `WebSocket.connect`
            and `WebSocket.disconnect` methods.
        """
        loop = asyncio.get_event_loop()

        def handle_msg(raw_msg: str) -> None:
            try:
                data: Dict[str, Any] = self._json_decoder.decode(raw_msg)
            except JSONDecodeError as e:
                log.error(f"Couldn't parse received JSON data in {self}: {e}\nmsg: {raw_msg}")
                return

            if not isinstance(data, dict):
                log.warning(f"Received invalid message type in {self}. "
                            f"Expecting object, received type {type(data).__name__}: {data}")
                return

            _ = self.event_target.emit(RawMsgReceiveEvent(self, data))

            try:
                op = data.pop("op")
            except KeyError:
                log.info(f"Ignoring message without op code in {self}: {data}")
                return

            event_type = data.get("type")
            cls = andesite.get_update_model(op, event_type)
            if cls is None:
                log.warning(f"Ignoring message with unknown op \"{op}\" in {self}: {data}")
                return

            try:
                message: andesite.ReceiveOperation = build_from_raw(cls, data)
            except Exception:
                log.exception(f"Couldn't parse message in {self} from Andesite node to {cls}: {data}")
                return

            message.client = self

            if isinstance(message, andesite.ConnectionUpdate):
                log.info(f"received connection update, setting last connection id in {self}.")
                self.__last_connection_id = message.id

            _ = self.event_target.emit(message)

            if self.state is not None:
                loop.create_task(self.state._handle_andesite_message(message))

        while True:
            try:
                raw_msg = await self.web_socket_client.recv()
            except asyncio.CancelledError:
                break
            except ConnectionClosed:
                log.error(f"Disconnected from websocket, trying to reconnect {self}!")
                _ = self.event_target.emit(WebSocketDisconnectEvent(self, False))
                await self.connect()
                continue

            if log.isEnabledFor(logging.DEBUG):
                log.debug(f"Received message in {self}: {raw_msg}")

            try:
                handle_msg(raw_msg)
            except Exception:
                log.exception("Exception in %s while handling message %s.", self, raw_msg)

    def _start_read_loop(self, *, loop: asyncio.AbstractEventLoop = None) -> None:
        """Start the web socket reader.

        If the reader is already running, this is a no-op.
        """
        if self.__read_loop and not self.__read_loop.done():
            return

        if loop is None:
            loop = asyncio.get_event_loop()

        self.__read_loop = loop.create_task(self._web_socket_reader())

    def _stop_read_loop(self) -> None:
        """Stop the web socket reader.

        If the reader is already stopped, this is a no-op.
        """
        if not self.__read_loop:
            return

        self.__read_loop.cancel()

    async def send(self, guild_id: int, op: str, payload: Dict[str, Any]) -> None:
        payload.update(guildId=str(guild_id), op=op)

        _ = self.event_target.emit(RawMsgSendEvent(self, guild_id, op, payload))

        if log.isEnabledFor(logging.DEBUG):
            log.debug(f"sending payload: {payload}")

        data = self._json_encoder.encode(payload)

        if self.web_socket_client is not None:
            try:
                await self.web_socket_client.send(data)
            except ConnectionClosed:
                # let the websocket reader handle this
                pass
            else:
                state = self.state
                if state:
                    _ = state._handle_sent_message(guild_id, op, payload)

                return

        log.info("Not connected, adding message to queue.")
        self.__message_queue.append(data)
        await self.connect()


class WebSocket(WebSocketBase, aiobservable.Observable, WebSocketInterface):
    ...



"""Client pools for multiple clients.

Client pools work exactly like the other clients,
but internally they use more than one client
(preferably to multiple Andesite nodes).

Attributes:
    RegionGuildComparator (Callable[[int, Optional[str]], int]: (Type alias)
        Function which takes a guild id and the node region of an Andesite node and
        returns an integer to indicate how well the guild is suited for the region.
    PoolScoringFunction (Callable[[ScoringData], Union[Any, Awaitable[Any]]]): (Type alias)
        Function which takes `ScoringData` as its only argument and returns a comparable object.
        The function may also be a coroutine. The return value is only compared to return values
        of the same function and the comparisons are bigger than (>) and less than (<). Equality (==)
        is implied if something is neither bigger nor smaller than the other value.
        A bigger return value (by comparison) implies that the `ScoringData` is better
        (ex: more suitable for the given guild).
    NodeDetails (Tuple[Union[str, yarl.URL], Optional[str]]): (Type alias) Tuple
        containing the uri and the password of an Andesite node.
"""

import abc
import asyncio
import dataclasses
import inspect
import logging
import random
import time
from collections import defaultdict, deque
from contextlib import suppress
from typing import Any, Awaitable, Callable, Collection, Deque, Dict, Generic, Iterable, Iterator, List, Mapping, \
    MutableMapping, Optional, Set, Tuple, TypeVar, Union
from weakref import WeakKeyDictionary, WeakValueDictionary

import aiobservable
import yarl

import andesite

__all__ = ["PoolException", "PoolEmptyError",
           "PoolClientAddEvent", "PoolClientRemoveEvent",
           "ClientPool",
           "HTTPPoolBase", "HTTPPool",
           "RegionGuildComparator", "ScoringData", "PoolScoringFunction", "default_scoring_function",
           "WebSocketPoolBase", "WebSocketPool",
           "create_pool"]

log = logging.getLogger(__name__)

CT = TypeVar("CT")


class PoolException(Exception):
    """Pool exceptions.

    Attributes:
        pool (ClientPool): Pool which raised the error
    """
    __slots__ = ("pool",)

    pool: "ClientPool"

    def __init__(self, pool: "ClientPool") -> None:
        super().__init__()
        self.pool = pool


class PoolEmptyError(PoolException):
    """Raised when a pool is empty but shouldn't be."""
    __slots__ = ()


@dataclasses.dataclass()
class PoolClientAddEvent(Generic[CT]):
    """Event for when a new client is added to a pool.

    Attributes:
        pool (ClientPool): Pool the client was added to
        client (CT): Client that was added to the pool
    """
    pool: "ClientPool"
    client: CT


@dataclasses.dataclass()
class PoolClientRemoveEvent(Generic[CT]):
    """Event for when a client is removed from a pool.

    Attributes:
        pool (ClientPool): Pool the client was added to
        client (CT): Client that was added to the pool
    """
    pool: "ClientPool"
    client: CT


class ClientPool(aiobservable.Observable, Collection[CT], abc.ABC, Generic[CT]):
    """Andesite client pool."""
    _clients: Collection[CT]

    def __repr__(self) -> str:
        args_str = ", ".join(repr(client) for client in self._clients)
        return f"{type(self).__name__}({args_str})"

    def __str__(self) -> str:
        return f"{type(self).__name__} [{len(self)} clients]"

    def __len__(self) -> int:
        return len(self._clients)

    def __bool__(self) -> bool:
        return not self.closed

    def __contains__(self, client: CT) -> bool:
        return client in self._clients

    def __iter__(self) -> Iterator[CT]:
        return iter(self._clients)

    @property
    def closed(self) -> bool:
        """Check whether all clients in the pool are closed."""
        return all(client.closed for client in self._clients)

    async def close(self) -> None:
        """Close all clients in the pool."""
        await asyncio.gather(*(client.close() for client in self._clients))

    async def reset(self) -> None:
        """Reset all underlying clients so they may be used again.

        This has the opposite effect of the `close` method making the clients
        usable again.
        """
        await asyncio.gather(*(client.reset() for client in self))

    def _check_client(self, client: CT) -> Optional[CT]:
        """Check whether the client is usable.

        Depending on the situation the client might even
        get removed from the pool entirely, such as when
        it is closed.

        Returns:
            Provided client if usable, otherwise `None`.
        """
        if client.closed:
            log.info(f"{client} closed, removing from pool {self}")
            self.remove_client(client)
        else:
            return client

        return None

    @abc.abstractmethod
    def add_client(self, client: CT) -> None:
        """Add a new client to the pool.

        Args:
            client: Client to add.

        Raises:
            ValueError: If the client is already in the pool

        Dispatches the `PoolClientAddEvent` event.
        """
        if client in self:
            raise ValueError(f"Client {client!r} already in {self}")

        _ = self.emit(PoolClientAddEvent(self, client))

    @abc.abstractmethod
    def remove_client(self, client: CT) -> None:
        """Remove a client from the pool.

        Args:
            client: Client to remove.

        Raises:
            ValueError: If the client isn't in the pool

        Dispatches the `PoolClientRemoveEvent` event.
        """
        _ = self.emit(PoolClientRemoveEvent(self, client))


# HTTP pool

class HTTPPoolBase(ClientPool[andesite.AbstractHTTP], andesite.AbstractHTTP):
    """Andesite HTTP client pool.

    Args:
        clients: HTTP clients to initialise the pool with
        timeout: Sets the `timeout` attribute.
        max_penalties: Sets the `max_penalties` attribute.
        penalty_time_frame: Sets the `penalty_time_frame` attribute.

    The pool uses a circular buffer which is rotated after every request.

    Attributes:
        timeout: Time in seconds to wait before starting the request on the
            next client. Note that the previous request isn't cancelled, if it
            succeeds after the next attempt has been started it is still
            accepted and returned.
        max_penalties: Max number of penalties a client may receive in the
            `penalty_time_frame` before being removed from the pool. A penalty
            is added to a client each time it raises an unexpected error.
        penalty_time_frame: Number of seconds before a penalty expires and no
            long counts toward a client's total number of penalties.
    """
    _clients: Deque[andesite.AbstractHTTP]
    _penalties: Dict[andesite.AbstractHTTP, List[float]]

    timeout: float
    max_penalties: int
    penalty_time_frame: float

    def __init__(self, clients: Iterable[andesite.AbstractHTTP], *,
                 timeout: float = None,
                 max_penalties: int = 5,
                 penalty_time_frame: float = 60) -> None:
        super().__init__()

        # collect in a set first to make sure clients are unique.
        self._clients = deque(set(clients))
        self._penalties = defaultdict(list)

        self.timeout = timeout
        self.max_penalties = max_penalties
        self.penalty_time_frame = penalty_time_frame

    def add_client(self, client: andesite.AbstractHTTP) -> None:
        """Add a new client to the pool.

        Args:
            client: Client to add.

        Raises:
            ValueError: If the client is already in the pool
        """
        super().add_client(client)
        self._clients.append(client)

    def remove_client(self, client: andesite.AbstractHTTP) -> None:
        """Remove a client from the pool.

        Args:
            client: Client to remove.

        Raises:
            ValueError: If the client isn't in the pool
        """
        self._clients.remove(client)
        with suppress(KeyError):
            del self._penalties[client]

        super().remove_client(client)

    def get_current_client(self) -> Optional[andesite.AbstractHTTP]:
        """Get the current http client.

        Returns:
            The current client or `None` if there are no clients.
        """
        try:
            return self._clients[0]
        except IndexError:
            return None

    def get_next_client(self) -> Optional[andesite.AbstractHTTP]:
        """Move to the next http client and return it.

        In reality this returns the next client
        that seems to be working. (i.e. not closed)

        Returns:
            Next available client. `None` if no clients are available.
        """
        while True:
            self._clients.rotate(1)
            client = self.get_current_client()

            # no clients in pool
            if client is None:
                return None

            client = self._check_client(client)
            if client is None:
                continue

            # everything seems in order with this client
            return client

    def _add_penalty(self, client: andesite.AbstractHTTP) -> None:
        penalties = self._penalties[client]

        new_ts = time.monotonic()
        min_ts = new_ts - self.penalty_time_frame

        # find and remove all expired penalties
        i = 0
        for i, ts in enumerate(penalties):
            if ts >= min_ts:
                break

        del penalties[:i]

        penalties.append(new_ts)

        log.info(f"added penalty to {client} in {self}")

        if len(penalties) > self.max_penalties:
            log.warning(f"{client} has too many penalties, removing from {self}")
            self.remove_client(client)

    async def request(self, method: str, path: str, **kwargs) -> Any:
        """Perform a request on the one of the clients.

        See Also:
            `AbstractHTTP.request` for the documentation.

        Raises:
            PoolEmptyError: If no clients are available
        """
        running_fs: Set[asyncio.Future] = set()
        loop = asyncio.get_event_loop()

        while True:
            client = self.get_next_client()
            if client is None:
                raise PoolEmptyError(self)

            fut = loop.create_task(client.request(method, path, **kwargs))
            running_fs.add(fut)

            done_fs, pending_fs = await asyncio.wait(running_fs, return_when=asyncio.FIRST_COMPLETED)
            try:
                done_fut = done_fs.pop()
            except KeyError:
                log.info(f"Requests to {len(running_fs)} client(s) timed-out, starting next request")
                continue

            try:
                return done_fut.result()
            except andesite.HTTPError:
                raise
            except Exception as e:
                log.info(f"error during request in {self}: {e}")
                self._add_penalty(client)


class HTTPPool(HTTPPoolBase, andesite.HTTPInterface):
    """Andesite HTTP client pool.

    This is just a wrapper around `HTTPPoolBase` which adds the
    `HTTPInterface` methods.

    Please see the documentation of `HTTPPoolBase` for more details.
    """
    ...


# WebSocket pool

RegionGuildComparator = Callable[[int, Optional[str]], int]


@dataclasses.dataclass()
class ScoringData:
    """Data passed to the web socket scoring scoring function.

    Attributes:
        pool (WebSocketPoolBase):
        client (AbstractWebSocket): Client to be evaluated
        node_region (Optional[str]): Node region reported by the Andesite
            client.
        region_comparator (Optional[RegionGuildComparator]): Function to
            compare node region with guild id.
        guild_ids (Set[int]): Guild ids which are already assigned to the client
        guild_id (int): Guild id which the client should be evaluated for
    """
    pool: "WebSocketPoolBase"
    client: andesite.AbstractWebSocket
    node_region: Optional[str]
    region_comparator: Optional[RegionGuildComparator]
    guild_ids: Set[int]
    guild_id: int


PoolScoringFunction = Callable[[ScoringData], Union[Any, Awaitable[Any]]]


def default_scoring_function(data: ScoringData) -> Tuple[int, int, int]:
    """Calculate the default score.

    Returns:
        3-tuple with the first element representing the connection status
        -1 for closed, 0 for disconnected, and 1 for connected.
        If the client doesn't have a `connected` attribute it is still
        set to 1.

        The second element is the result of the region comparator or
        0 if no region comparator was passed.

        The third element is the amount of guilds negative.
        So it will favour clients with less guilds.
    """
    client = data.client
    if client.closed:
        return -1, 0, 0

    try:
        # noinspection PyUnresolvedReferences
        connected = client.connected
    except AttributeError:
        connected = 1
    else:
        connected = 1 if connected else 0

    if data.region_comparator:
        region_score = data.region_comparator(data.guild_id, data.node_region)
    else:
        region_score = 0

    return connected, region_score, -len(data.guild_ids)


class WebSocketPoolBase(ClientPool[andesite.AbstractWebSocket], andesite.AbstractWebSocket):
    """Base implementation of a web socket pool.

    Args:
        clients: Web socket clients to initialise the pool with.
        state: State handler to use for this pool. The state is used
            to migrate nodes and defaults to the in-memory `State`.
            You can however disable the state by passing `False`.
        scoring_function: Scoring function to evaluate clients with.
            This is a function which takes `ScoringData` as its only argument
            and returns a comparable object. The function can be a coroutine.
            Defaults to `default_scoring_function`.
        region_comparator: Region comparator to use for comparing regions.

    The pool uses different clients on a guild to guild level.

    If the pool uses a state, all clients within the pool are forced to use the
    same state as well. Trying to add a client with a state to a pool with a
    state will result in an error.
    """

    _clients: Set[andesite.AbstractWebSocket]

    _guild_clients: MutableMapping[int, andesite.AbstractWebSocket]
    _client_guilds: MutableMapping[andesite.AbstractWebSocket, Set[int]]

    _scoring_function: PoolScoringFunction
    _scoring_func_is_coro: bool
    _region_comparator: RegionGuildComparator

    def __init__(self, clients: Iterable[andesite.AbstractWebSocket], *,
                 state: andesite.StateArgumentType = None,
                 scoring_function: PoolScoringFunction = None,
                 region_comparator: RegionGuildComparator = None) -> None:
        super().__init__()

        self._clients = set()
        self._guild_clients = WeakValueDictionary()
        self._client_guilds = WeakKeyDictionary()

        self.state = state

        self.scoring_function = scoring_function or default_scoring_function
        self.region_comparator = region_comparator

        for client in clients:
            self.add_client(client)

    @property
    def state(self) -> Optional[andesite.AbstractState]:
        return super().state

    @state.setter
    def state(self, value: Optional[andesite.AbstractState]) -> None:
        super(WebSocketPoolBase, type(self)).state.fset(self, value)

        for client in self:
            client.state = value

    @property
    def scoring_function(self) -> PoolScoringFunction:
        """Scoring function used by the pool."""
        return self._scoring_function

    @scoring_function.setter
    def scoring_function(self, func: PoolScoringFunction) -> None:
        """Set the scoring function.

        Args:
            func: Scoring function to use.
        """
        self._scoring_function = func
        self._scoring_func_is_coro = inspect.iscoroutine(func)

    @property
    def region_comparator(self) -> Optional[RegionGuildComparator]:
        """Region comparator used to compare guild region with Andesite node region."""
        return self._region_comparator

    @region_comparator.setter
    def region_comparator(self, value: Optional[RegionGuildComparator]) -> None:
        """Set the comparator used to compare guild region with Andesite node region.

        Args:
            value: New comparator to use.
        """
        self._region_comparator = value

    def add_client(self, client: andesite.AbstractWebSocket) -> None:
        """Add a new client to the pool.

        Args:
            client: Client to add.

        Raises:
            ValueError: If the client is already in the pool or if the client
                has a state and the pool also has a different state.
        """
        super().add_client(client)
        self._clients.add(client)
        self._client_guilds[client] = set()
        client.event_target.add_child(self.event_target)

        state = self.state
        if state is not None:
            client_state = client.state
            if client_state is not None and client_state is not state:
                raise ValueError(f"Cannot add client {client} with state to pool {self} with state")

            client.state = state

    def remove_client(self, client: andesite.AbstractWebSocket) -> None:
        """Remove a web socket client from the pool.

        Args:
            client: WebSocket client to remove

        Raises:
            ValueError: If the client isn't in the pool

        Notes:
            This method removes the client without performing node migration.
            If you want to properly remove a client use the `pull_client`
            method.
        """
        try:
            self._clients.remove(client)
        except KeyError:
            raise ValueError(f"Cannot remove {client!r}, not in {self}!")

        guild_ids = self._client_guilds.pop(client)

        for guild_id in guild_ids:
            del self._guild_clients[guild_id]

        # either both None in which case it doesn't matter or both the same state
        if client.state is self.state:
            client.state = False

        client.event_target.remove_child(self.event_target)

        super().remove_client(client)

    async def pull_client(self, client: andesite.AbstractWebSocket) -> None:
        """Remove a client from the pool and migrate its state to another node.

        Args:
            client: WebSocket client to remove

        Raises:
            ValueError: If the client isn't in the pool
        """
        guild_ids = self.get_guild_ids(client)
        self.remove_client(client)

        # TODO destroy all players

        # TODO transfer message queue

        fs: List[asyncio.Future] = []
        loop = asyncio.get_event_loop()

        for guild_id in guild_ids:
            fut = loop.create_task(self.assign_client(guild_id))
            fs.append(fut)

        new_clients: List[Optional[andesite.AbstractWebSocket]] = await asyncio.gather(*fs)

        async def load_player_state(_guild_id: int, _client: andesite.AbstractWebSocket) -> None:
            player_state = await self.state.get(_guild_id)
            await _client.load_player_state(player_state)

        fs.clear()

        for guild_id, new_client in zip(guild_ids, new_clients):
            fut = loop.create_task(load_player_state(guild_id, new_client))
            fs.append(fut)

        await asyncio.gather(*fs)

    def get_client(self, guild_id: int) -> Optional[andesite.AbstractWebSocket]:
        """Get the andesite web socket client which is used for the given guild."""
        try:
            client = self._guild_clients[guild_id]
        except KeyError:
            client = None
        else:
            client = self._check_client(client)

        return client

    def iter_client_guild_ids(self) -> Iterator[Tuple[andesite.AbstractWebSocket, Set[int]]]:
        """Iterate over all clients with their assigned guild ids."""
        for client in self._clients:
            yield client, self.get_guild_ids(client)

    async def calculate_score(self, data: ScoringData) -> Any:
        """Calculate the score using the scoring function."""
        if self._scoring_func_is_coro:
            return await self._scoring_function(data)
        else:
            return self._scoring_function(data)

    async def find_best_client(self, guild_id: int) -> Optional[andesite.AbstractWebSocket]:
        """Determine the best client for the given guild.

        If there are multiple clients with the same score, a
        random one is returned.
        """
        best_clients: List[andesite.AbstractWebSocket] = []
        best_score: Any = None

        score_fs: List[asyncio.Future] = []

        async def perform_score_calc(_score_data: ScoringData) -> Tuple[andesite.AbstractWebSocket, Any]:
            return _score_data.client, await self.calculate_score(_score_data)

        loop = asyncio.get_event_loop()

        for client, guild_ids in self.iter_client_guild_ids():
            try:
                # noinspection PyUnresolvedReferences
                node_region: Optional[str] = client.node_region
            except AttributeError:
                node_region = None

            score_data = ScoringData(self, client, node_region, self.region_comparator, guild_ids, guild_id)
            score_fs.append(loop.create_task(perform_score_calc(score_data)))

        scores: List[Tuple[andesite.AbstractWebSocket, Any]] = await asyncio.gather(*score_fs)
        for client, score in scores:
            if best_score is not None:
                if score < best_score:
                    continue
                elif score > best_score:
                    best_clients = []

            best_score = score
            best_clients.append(client)

        try:
            return random.choice(best_clients)
        except IndexError:
            return None

    def _assign_client(self, client: andesite.AbstractWebSocket, guild_id: int) -> None:
        """Internally assign a guild to a client.

        This creates both the guild id -> client and
        the client -> guild ids relation.

        Args:
            client: Client to assign to the guild id
            guild_id: Guild to assign it to
        """
        self._guild_clients[guild_id] = client

        try:
            guild_ids = self._client_guilds[client]
        except KeyError:
            guild_ids = self._client_guilds[client] = set()

        guild_ids.add(guild_id)

    async def assign_client(self, guild_id: int) -> Optional[andesite.AbstractWebSocket]:
        """Assign a client to the given guild.

        If the guild already has a client, it is simply returned.
        """
        client = self.get_client(guild_id)
        if client is None:
            client = await self.find_best_client(guild_id)
            self._assign_client(client, guild_id)

        return client

    def get_guild_ids(self, client: andesite.AbstractWebSocket) -> Set[int]:
        """Get the guild ids which the client is assigned to.

        Args:
            client: Client to get guild ids for.

        Returns:
            Guild ids set which are assigned to the client.
            Mutating the returned set won't affect the internal
            state, the returned set is a shallow copy of the internal one.
        """
        try:
            return self._client_guilds[client].copy()
        except KeyError:
            return set()

    async def send(self, guild_id: int, op: str, payload: Dict[str, Any]) -> None:
        # assign client returns the client if one is already assigned
        client = await self.assign_client(guild_id)

        if not client:
            raise PoolEmptyError(self)

        await client.send(guild_id, op, payload)


class WebSocketPool(WebSocketPoolBase, andesite.WebSocketInterface):
    """Pool of Andesite web socket connections.

    Please refer to `WebSocketPoolBase` for the documentation.
    """
    ...


NodeDetails = Tuple[Union[str, yarl.URL], Optional[str]]


def create_pool(http_nodes: Iterable[NodeDetails],
                web_socket_nodes: Iterable[NodeDetails], *,
                user_id: int,
                state: andesite.StateArgumentType = None,
                http_pool_kwargs: Mapping[str, Any] = None,
                web_socket_pool_kwargs: Mapping[str, Any] = None) -> andesite.Client:
    """Create an `Client` with client pools.

    Uses `HTTPBase` and `WebSocketBase` which are contained in
    `HTTPPoolBase` and `WebSocketPoolBase` pools respectively.

    Args:
        http_nodes: Tuples of [uri, password] for each REST node to connect to.
        web_socket_nodes: Tuples of [uri, password] for each WebSocket node to
            connect to.
        user_id: Bot's user id.
        state: State handler to use for the pools. Defaults to `State`,
            because a state handler is required for node migration to work. You
            can pass `False` to disable state handling though.
        http_pool_kwargs: Additional keyword arguments to pass to the http pool
            constructor.
        web_socket_pool_kwargs: Additional keyword arguments to pass to the web
            socket pool constructor.

    Returns:
        A combined Andesite client operation on an http pool and a web socket
        pool.
    """
    http_clients = []
    for uri, password in http_nodes:
        http_clients.append(andesite.HTTPBase(uri, password))

    web_socket_clients = []
    for uri, password in web_socket_nodes:
        ws = andesite.WebSocketBase(uri, user_id, password, state=False)
        web_socket_clients.append(ws)

    http_pool_kwargs = http_pool_kwargs or {}
    http_pool = HTTPPoolBase(http_clients, **http_pool_kwargs)

    web_socket_pool_kwargs = web_socket_pool_kwargs or {}
    web_socket_pool = WebSocketPoolBase(web_socket_clients, **web_socket_pool_kwargs)

    inst = andesite.Client(http_pool, web_socket_pool)

    # the setter will make sure the state is proper
    inst.state = state
    return inst



"""Utilities for the discord.py library.

If you're using the discord.py library, you can use the functions provided by
this module to your advantage. Instead of having to manually send the voice
server updates, you can call `add_voice_server_update_handler` once and be done
with it. All events received by discord.py will then automatically be forwarded
to the Andesite client.
Even further, you can use `connect_voice_channel` and `disconnect_voice_channel`
to easily connect to and disconnect from voice channels.
This works for both `discord.ext.commands.Bot` and normal `discord.Client`
instances without interfering.

Attributes:
    SOCKET_RESPONSE_HANDLERS_ATTR (str): Name of the attribute used to store the
        `SocketResponseHandler` instances in discord clients.
"""
import asyncio
import logging
from asyncio import Future
from typing import Any, Callable, Dict, Iterable, Optional, Set, TYPE_CHECKING, Tuple, Union, cast, overload

import andesite

if TYPE_CHECKING:
    try:
        from discord import Client, VoiceChannel, Guild
        # noinspection PyUnresolvedReferences
        from discord.ext.commands import Bot
        from discord.gateway import DiscordWebSocket
        from discord.state import ConnectionState
    except ImportError:
        pass

__all__ = ["get_discord_websocket",
           "update_voice_state", "connect_voice_channel", "disconnect_voice_channel",
           "AsyncMethodGroup", "get_async_method_group", "wrap_client_listener", "unwrap_client_listener",
           "SocketResponseHandler", "get_andesite_socket_response_handlers",
           "add_voice_server_update_handler", "remove_voice_server_update_handler",
           "compare_regions", "create_region_comparator"]

log = logging.getLogger(__name__)


def get_discord_websocket(client: Union["Client", "ConnectionState"], guild_id: int) -> "DiscordWebSocket":
    """Utility method to get access to discord.py's gateway websocket.

    Args:
        client: discord.py client.
        guild_id: Guild id whose websocket to get.
    """
    try:
        # noinspection PyProtectedMember
        getter = client._connection._get_websocket
    except AttributeError:
        # noinspection PyProtectedMember,PyUnresolvedReferences
        getter = client._get_websocket

    return getter(guild_id)


async def update_voice_state(client: "Client", guild_id: int, channel_id: Optional[int]) -> None:
    """Update the voice state.

    Args:
        client: discord.py client.
        guild_id: Guild id to target
        channel_id: Channel id to connect to.
            If `None`, disconnect from the current channel.
    """
    ws = get_discord_websocket(client, guild_id)
    await ws.voice_state(guild_id, channel_id)


@overload
async def connect_voice_channel(client: "Client", channel: "VoiceChannel") -> None: ...


@overload
async def connect_voice_channel(client: "Client", guild: Union["Guild", int], channel_id: int) -> None: ...


async def connect_voice_channel(client: "Client", *args, **kwargs) -> None:
    """Connect to a voice channel.

    This function has two signatures, you can either call it with a `VoiceChannel`,
    or provide a guild / guild id and the voice channel id.
    """
    total_args_len = len(args) + len(kwargs)

    # use this rather peculiar method to allow **kwargs and *args interchangeably
    if total_args_len == 1:
        await _connect_voice_channel_channel(client, *args, **kwargs)
    else:
        await _connect_voice_channel_guild_channel(client, *args, **kwargs)


async def _connect_voice_channel_channel(client: "Client", channel: "VoiceChannel") -> None:
    await update_voice_state(client, channel.guild.id, channel.id)


async def _connect_voice_channel_guild_channel(client: "Client", guild: Union["Guild", int], channel_id: int) -> None:
    guild_id = guild if isinstance(guild, int) else guild.id
    await update_voice_state(client, guild_id, channel_id)


async def disconnect_voice_channel(client: "Client", guild: Union["Guild", int]) -> None:
    """Disconnect from the current voice channel."""
    guild_id = guild if isinstance(guild, int) else guild.id
    await update_voice_state(client, guild_id, None)


class AsyncMethodGroup:
    """Group of async functions which act as a method.

    When called the instance calls all its functions.
    It doesn't pass the "self" parameter.
    """
    __slots__ = ("methods",)

    methods: Set[Callable]

    def __init__(self, methods: Iterable[Callable]) -> None:
        self.methods = set(methods)

    def __call__(self, obj: Any, *args, **kwargs) -> Future:
        return asyncio.gather(func(*args, **kwargs) for func in self.methods)


def get_async_method_group(obj: Any, name: str) -> AsyncMethodGroup:
    """Get a method group for a method.

    If the method exists and is not already an `AsyncMethodGroup`,
    a new group containing the previous method takes its place.
    """
    try:
        method = getattr(obj, name)
    except AttributeError:
        method = AsyncMethodGroup([])
    else:
        if isinstance(method, AsyncMethodGroup):
            return method
        else:
            method = AsyncMethodGroup([method])

    setattr(obj, name, method)
    return method


def wrap_client_listener(client: "Client", func: Callable, *, name: str = None) -> None:
    """Add a listener method to the discord client.

    Args:
        client: Client to add the listener to
        func: Listener function to add
        name: Custom name to use. Defaults to the name of the function.
            Be sure to include "on\_" if you set this.

    This makes it possible to add a listener if the client isn't a `discord.ext.commands.Bot`.
    This is achieved by adding the listener as a method to the client.
    If there already is a listener it is still called!

    See Also:
        `unwrap_client_listener` to remove it again.
    """
    name = name or func.__name__

    group = get_async_method_group(client, name)
    group.methods.add(func)


def unwrap_client_listener(client: "Client", func: Callable, *, name: str = None) -> None:
    """Remove a listener method from a discord client.

    Args:
        client: Client to remove listener from
        func: Listener function to remove
        name: Custom name to use. Defaults to the name of the function.
            Be sure to include "on\_" if you set this.
    """
    name = name or func.__name__

    group = get_async_method_group(client, name)
    group.methods.remove(func)


class SocketResponseHandler:
    """Socket response listener.

    An interface between discord.py clients and andesite clients to automatically
    send voice server updates.

    Attributes:
        discord_client (discord.Client): discord.py client that is listened to.
        andesite_client (WebSocketInterface): Andesite client to send
            the voice server update.
    """
    __slots__ = ("discord_client", "andesite_client")

    discord_client: "Client"
    andesite_client: andesite.WebSocketInterface

    def __init__(self, discord_client: "Client", andesite_client: andesite.WebSocketInterface) -> None:
        self.discord_client = discord_client
        self.andesite_client = andesite_client

    def add_listener(self) -> None:
        """Add the on_socket_response listener.

        If the handler is attached to a `Bot`, it uses the listener framework,
        otherwise it safely wraps the client handler.
        """
        client = self.discord_client

        try:
            add_listener = cast("Bot", client).add_listener
        except AttributeError:
            wrap_client_listener(self.discord_client, self.on_socket_response)
        else:
            log.info(f"Adding socket response listener to {client}")
            add_listener(self.on_socket_response)

    def remove_listener(self) -> None:
        """Remove the on_socket_response listener."""
        client = self.discord_client

        try:
            remove_listener = cast("Bot", client).remove_listener
        except AttributeError:
            unwrap_client_listener(self.discord_client, self.on_socket_response)
        else:
            log.info(f"Removing socket response listener from {client}")
            remove_listener(self.on_socket_response)

    async def on_socket_response(self, data: Dict[str, Any]) -> None:
        """Intercept voice server updates and send them to Andesite."""
        try:
            key = data["t"]
            body = data["d"]
        except KeyError:
            return

        if key != "VOICE_SERVER_UPDATE":
            return

        guild_id = int(body["guild_id"])
        ws = get_discord_websocket(self.discord_client, guild_id)
        session_id = ws.session_id

        if session_id:
            log.info(f"sending voice server update for guild {guild_id}")
            await self.andesite_client.voice_server_update(guild_id, session_id, body)
        else:
            log.debug(f"not sending voice server update for guild {guild_id} because session id missing.")


SOCKET_RESPONSE_HANDLERS_ATTR: str = "__andesite_socket_response_handlers__"


def get_andesite_socket_response_handlers(obj: Any) -> Dict[andesite.WebSocketInterface, SocketResponseHandler]:
    """Get the socket response handlers added to the discord client.

    If it doesn't exist, a new one is created and added to the object.
    """
    try:
        handlers = getattr(obj, SOCKET_RESPONSE_HANDLERS_ATTR)
    except AttributeError:
        handlers = {}
        setattr(obj, SOCKET_RESPONSE_HANDLERS_ATTR, handlers)

    return handlers


def add_voice_server_update_handler(discord_client: "Client", andesite_client: andesite.WebSocketInterface) -> None:
    """Add a voice server update listener to the discord client.

    Args:
        discord_client: Discord client to add the listener to.
        andesite_client: Andesite web socket client to use to send the voice server update.

    This will listen to socket responses using the discord client
    and trigger `WebSocketInterface.voice_server_update`.
    """
    handlers = get_andesite_socket_response_handlers(discord_client)

    if andesite_client in handlers:
        return

    handler = SocketResponseHandler(discord_client, andesite_client)
    handler.add_listener()

    handlers[andesite_client] = handler


def remove_voice_server_update_handler(discord_client: "Client", andesite_client: andesite.WebSocketInterface) -> None:
    """Remove the socket response handler added by `add_voice_server_update_handler`.

    Args:
        discord_client: Discord client to remove listener from.
        andesite_client: Andesite web socket client to use to send the voice server update.
    """
    handlers = get_andesite_socket_response_handlers(discord_client)

    try:
        handler = handlers.pop(andesite_client)
    except KeyError:
        return

    handler.remove_listener()


def _split_region(region: str) -> Tuple[bool, str, Optional[str]]:
    """Parse a region string into its parts.

    Args:
        region: Region string to parse

    Returns:
        3-tuple:

            - Whether the region is VIP or not
            - Country name
            - Part of country
    """
    parts = region.lower().split("-")

    if parts[0] == "vip":
        del parts[0]
        vip = True
    else:
        vip = False

    country = parts.pop(0)

    try:
        country_part = parts.pop(0)
    except IndexError:
        country_part = None

    return vip, country, country_part


def compare_regions(a_region: str, b_region: str) -> int:
    """Compare two region names.

    The order of the provided regions is irrelevant.

    Args:
        a_region: First region name
        b_region: Second region name

    Returns:
        0 if the regions can't be properly compared
        (ex: Node region unknown or guild id unknown).
        If the regions can be compared
        the result is the sum of the following points:

        - 2 points if both regions are in the same country
        - 1 point if both regions are in
          the same part of a country (ex: us_west)
          (This point is also awarded if both regions
          don't specify a country part)
    """
    score: int = 0

    _, a_country, a_country_part = _split_region(a_region)
    _, b_country, b_country_part = _split_region(b_region)

    if a_country == b_country:
        score += 2

        if a_country_part == b_country_part:
            score += 1

    return score


def create_region_comparator(discord_client: "Client", *,
                             region_comp: Callable[[str, str], int] = None) -> andesite.RegionGuildComparator:
    """Create a region comparator which compares the guild region with the node region.

    You can use the created comparator for the `WebSocketPool`.

    Args:
        discord_client: Client to use to determine the guild region.
        region_comp: Specify the actual function which compares the guild region
            and the Andesite node region. Defaults to `compare_regions`.

    Returns:
        Region comparator which uses region_comp to compare guild region with node region.
    """
    if region_comp is None:
        region_comp = compare_regions

    def comparator(guild_id: int, node_region: Optional[str]) -> int:
        if node_region is None or node_region == "unknown":
            return 0

        guild: Optional[Guild] = discord_client.get_guild(guild_id)
        if guild is None:
            return 0

        return region_comp(node_region, guild.region)

    return comparator



"""Operations sent by Andesite.

Attributes:
    EVENT_MAP (Mapping[str, Type[AndesiteEvent]]): Mapping from the event name
        to the corresponding `AndesiteEvent` type. See: `get_event_model`
    OP_MAP (Mapping[str, Type[ReceiveOperation]]): Mapping from the op code to
    the corresponding `ReceiveOperation` type. See: `get_update_model`

See Also:
    `andesite.models.send_operations` for operations sent to Andesite.
"""

from __future__ import annotations

import abc
import copy
import dataclasses
from dataclasses import dataclass
from enum import Enum
from typing import Dict, List, Mapping, Optional, Set, Tuple, Type, Union, cast

import andesite
from andesite.transform import RawDataType, map_build_values_from_raw, map_convert_values, \
    map_convert_values_from_milli, map_convert_values_to_milli, map_remove_keys, transform_input, transform_output
from .debug import Error, Stats
from .player import Player

__all__ = ["ReceiveOperation",
           "PongResponse",
           "ConnectionUpdate", "MetadataUpdate", "StatsUpdate", "PlayerUpdate",
           "AndesiteEvent", "TrackStartEvent",
           "TrackEndReason", "TrackEndEvent",
           "TrackExceptionEvent", "TrackStuckEvent",
           "WebSocketClosedEvent",
           "UnknownAndesiteEvent",
           "get_event_model", "get_update_model"]


class ReceiveOperation(abc.ABC):
    """Message sent by Andesite.

    Attributes:
        client (Optional[andesite.AbstractWebSocketClient]): Client that received
            the message. This is set by the client that received the message.
    """
    __op__: str

    client: Optional[andesite.AbstractWebSocketClient] = None


@dataclass
class PongResponse(ReceiveOperation):
    """Simple pong response sent as a response to ping requests.

    Attributes:
        user_id (int): User id
        guild_id (int): Guild id
    """
    __op__ = "pong"

    user_id: int
    guild_id: int

    @classmethod
    def __transform_input__(cls, data: RawDataType) -> None:
        map_convert_values(data, user_id=int, guild_id=int)

    @classmethod
    def __transform_output__(cls, data: RawDataType) -> None:
        map_convert_values(data, user_id=str, guild_id=str)


@dataclass
class ConnectionUpdate(ReceiveOperation):
    """Message sent upon connecting to the Web socket.

    Attributes:
        id (str): Connection ID
    """
    __op__ = "connection-id"

    id: str


@dataclass
class MetadataUpdate(ReceiveOperation):
    """Payload sent on connection start containing handshake response header.

    Attributes:
        data (Dict[str, Union[int, str, List[str]]]): Map of metadata key to
            value.
    """
    __op__ = "metadata"

    data: Dict[str, Union[int, str, List[str]]]


@dataclass
class StatsUpdate(ReceiveOperation):
    """Message containing statistics.

    Attributes:
        user_id (int): User ID
        stats (andesite.Stats): Statistics
    """
    __op__ = "stats"

    user_id: int
    stats: Stats

    @classmethod
    def __transform_input__(cls, data: RawDataType) -> None:
        map_convert_values(data, user_id=int)
        map_build_values_from_raw(data, stats=Stats)

    @classmethod
    def __transform_output__(cls, data: RawDataType) -> None:
        map_convert_values(data, user_id=str)


@dataclass
class PlayerUpdate(ReceiveOperation):
    """Player update sent by Andesite for active players.

    Attributes:
        user_id (int): user id
        guild_id (int): guild id
        state (andesite.Player): player
    """
    __op__ = "player-update"

    user_id: int
    guild_id: int
    state: Player

    @classmethod
    def __transform_input__(cls, data: RawDataType) -> None:
        map_convert_values(data, user_id=int, guild_id=int)
        map_build_values_from_raw(data, state=Player)

    @classmethod
    def __transform_output__(cls, data: RawDataType) -> None:
        map_convert_values(data, user_id=str, guild_id=str)


@dataclass
class AndesiteEvent(ReceiveOperation, abc.ABC):
    """Event sent by Andesite.

    Attributes:
        type (str): Event type name.
            This is equal to the name of the class (With the exception of
            `UnknownAndesiteEvent`)
        user_id (int): User ID
        guild_id (int): Guild ID
        track (str): Base64 encoded track data
    """
    __op__ = "event"

    type: str
    user_id: int
    guild_id: int

    @classmethod
    def __transform_input__(cls, data: RawDataType) -> None:
        map_convert_values(data, user_id=int, guild_id=int)

    @classmethod
    def __transform_output__(cls, data: RawDataType) -> None:
        map_convert_values(data, user_id=str, guild_id=str)


@dataclass
class TrackStartEvent(AndesiteEvent):
    """Event emitted when a new track starts playing."""

    track: str


class TrackEndReason(Enum):
    """Reason why a track stopped playing.

    See Also:
        `TrackEndEvent`

    Attributes:
        FINISHED: Usually caused by the track reaching the end,
            however it will also be used when it ends due to an exception.
        LOAD_FAILED: Track failed to start, throwing an exception before
            providing any audio.
        STOPPED: Track was stopped due to the player being stopped.
        REPLACED: Track stopped playing because a new track started playing.
        CLEANUP: Track was stopped because the cleanup threshold for the audio
            player was reached.
    """
    FINISHED = "FINISHED"
    LOAD_FAILED = "LOAD_FAILED"
    STOPPED = "STOPPED"
    REPLACED = "REPLACED"
    CLEANUP = "CLEANUP"


@dataclass
class TrackEndEvent(AndesiteEvent):
    """Event emitted when a track ended.

    Attributes:
        reason (TrackEndReason): Reason why a track stopped playing.
        may_start_next (bool): Indicates whether a new track should be started
            on receiving this event. If this is `False`, either this event is
            already triggered because another track started
            (`TrackEndReason.REPLACED`) or because the player is stopped
            (`TrackEndReason.STOPPED`, `TrackEndReason.CLEANUP`).
    """

    track: str
    reason: TrackEndReason
    may_start_next: bool

    @classmethod
    def __transform_input__(cls, data: RawDataType) -> RawDataType:
        data = transform_input(super(), data)
        map_convert_values(data, reason=TrackEndReason)
        return data

    @classmethod
    def __transform_output__(cls, data: RawDataType) -> RawDataType:
        data = transform_output(super(), data)
        data["reason"] = cast(TrackEndReason, data["reason"]).value
        return data


@dataclass
class TrackExceptionEvent(AndesiteEvent):
    """Event emitted when there's an error.

    Attributes:
        error (str): Error message
        exception (Error): Error data
    """
    track: str

    error: str
    exception: Error


@dataclass
class TrackStuckEvent(AndesiteEvent):
    """Event emitted when a track is stuck.

    Attributes:
        threshold (float): Threshold in seconds
    """

    track: str
    threshold: float

    @classmethod
    def __transform_input__(cls, data: RawDataType) -> RawDataType:
        data = transform_input(super(), data)
        map_convert_values_from_milli(data, "threshold")
        return data

    @classmethod
    def __transform_output__(cls, data: RawDataType) -> RawDataType:
        data = transform_output(super(), data)
        map_convert_values_to_milli(data, "threshold")
        return data


@dataclass
class WebSocketClosedEvent(AndesiteEvent):
    """Event emitted when the Andesite node disconnects from a voice channel.

    Attributes:
        reason (str): Reason for the disconnect
        code (int): Error code
        by_remote (bool): Whether the disconnect was caused by the remote.
    """

    reason: str
    code: int
    by_remote: bool


@dataclass
class UnknownAndesiteEvent(AndesiteEvent):
    """Special kind of event for unknown events.

    This shouldn't occur at all unless the library is out-dated.

    Attributes:
        body (object): Entire event body sent by Andesite.
            Please note that the keys are in snake_case.

    """

    body: RawDataType

    @classmethod
    def __transform_input__(cls, data: RawDataType) -> RawDataType:
        # we want a clean body... does that sound weird? no, I'm sure it doesn't
        data["body"] = copy.deepcopy(data)

        map_remove_keys(data, *(key for key in data.keys() if key not in _UNKNOWN_ANDESITE_EVENT_FIELD_NAMES))
        data = transform_input(super(), data)
        return data

    @classmethod
    def __transform_output__(cls, data: RawDataType) -> RawDataType:
        data = transform_output(super(), data)

        body: RawDataType = data.pop("body")
        # let the existing data overwrite the body
        body.update(data)
        return body


_UNKNOWN_ANDESITE_EVENT_FIELDS: Tuple[dataclasses.Field] = dataclasses.fields(UnknownAndesiteEvent)
_UNKNOWN_ANDESITE_EVENT_FIELD_NAMES: Set[str] = {field.name for field in _UNKNOWN_ANDESITE_EVENT_FIELDS}

_EVENTS: Set[Type[AndesiteEvent]] = {
    TrackStartEvent,
    TrackEndEvent,
    TrackExceptionEvent, TrackStuckEvent,
    WebSocketClosedEvent,
}
EVENT_MAP: Mapping[str, Type[AndesiteEvent]] = {event.__name__: event for event in _EVENTS}


def get_event_model(event_type: str) -> Type[AndesiteEvent]:
    """Get the model corresponding to the given event name.

    Args:
        event_type: Event type name.

    Returns:
        Model type for the given event type.
        `UnknownAndesiteEvent` if no matching event was found.
    """
    try:
        return EVENT_MAP[event_type]
    except KeyError:
        return UnknownAndesiteEvent


_OPS: Set[Type[ReceiveOperation]] = {PongResponse, ConnectionUpdate, MetadataUpdate, StatsUpdate, PlayerUpdate}
OP_MAP: Mapping[str, Type[ReceiveOperation]] = {op.__op__: op for op in _OPS}


def get_update_model(op: str, event_type: str = None) -> Optional[Type[ReceiveOperation]]:
    """Get the model corresponding to the given op code.

    Args:
        op: Op code sent by Andesite
        event_type: Event type if and only if op is "event".
            This is used to return the correct event type.
            See `get_event_model`.
            If not set and op is "event" the function returns `None`.
    """
    if op == "event":
        if event_type is None:
            return None

        return get_event_model(event_type)
    else:
        return OP_MAP.get(op)



"""A client library for Andesite.

The top level `andesite` namespace also
exports all the models from `andesite.models`
and the various clients.

Attributes:
    __version__: Version of the installed andesite.py library.
"""

# must be defined before http client is imported
__version__ = "0.2.1"

# preserve order:
from .models import *  # 1

from .http_client import *  # 2

from .state import *  # 3.1
from .web_socket_client import *  # 3.2
from .web_socket_client_events import *  # 3.3

from .combined_client import *  # 4
from .pools import *  # 5



from pathlib import Path

import setuptools

import andesite

long_description = Path("README.md").read_text()

setuptools.setup(
    name="andesite.py",
    version=andesite.__version__,
    description="Andesite client for Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Giesela Inc.",
    author_email="team@giesela.dev",
    url="https://github.com/gieseladev/andesite.py",

    packages=setuptools.find_packages(exclude=("examples", "docs", "tests")),
    python_requires="~=3.7",

    install_requires=[
        "aiobservable",
        "aiohttp",
        "lettercase",
        "websockets",
        "yarl",
    ],
)
